@@ -10,7 +10,7 @@
  *
  * Copyright (C) 2004 - 2005 Tresys Technology, LLC
  *	This program is free software; you can redistribute it and/or modify
- *	it under the terms of the GNU General Public License as published by
+ *  	it under the terms of the GNU General Public License as published by
  *	the Free Software Foundation, version 2.
  */
 
@@ -91,7 +91,7 @@ int define_policy(int pass, int module_header_given)
 		}
 	}
 	/* the first declaration within the global avrule
-	 * block will always have an id of 1 */
+	   block will always have an id of 1 */
 	next_decl_id = 2;
 
 	/* reset the scoping stack */
@@ -137,11 +137,16 @@ int declare_symbol(uint32_t symbol_type, hashtab_key_t key, hashtab_datum_t datu
 		return -1;
 	}
 	retval = symtab_insert(policydbp, symbol_type, key, datum, SCOPE_DECL, decl->decl_id, dest_value);
-	if (retval == 1) {
+	if (retval == 1 && dest_value) {
 		symtab_datum_t *s = (symtab_datum_t *) hashtab_search(policydbp->symtab[symbol_type].table,
 								      key);
 		assert(s != NULL);
-		*dest_value = s->value;
+
+		if (symbol_type == SYM_LEVELS) {
+			*dest_value = ((level_datum_t *) s)->level->sens;
+		} else {
+			*dest_value = s->value;
+		}
 	} else if (retval == -2) {
 		return -2;
 	} else if (retval < 0) {
@@ -156,6 +161,40 @@ int declare_symbol(uint32_t symbol_type, hashtab_key_t key, hashtab_datum_t datu
 	return retval;
 }
 
+static int role_implicit_bounds(hashtab_t roles_tab, char *role_id, role_datum_t * role)
+{
+	role_datum_t *bounds;
+	char *bounds_id, *delim;
+
+	delim = strrchr(role_id, '.');
+	if (!delim)
+		return 0;	       /* no implicit boundary */
+
+	bounds_id = strdup(role_id);
+	if (!bounds_id) {
+		yyerror("out of memory");
+		return -1;
+	}
+	bounds_id[(size_t) (delim - role_id)] = '\0';
+
+	bounds = hashtab_search(roles_tab, bounds_id);
+	if (!bounds) {
+		yyerror2("role %s doesn't exist, is implicit bounds of %s", bounds_id, role_id);
+		return -1;
+	}
+
+	if (!role->bounds)
+		role->bounds = bounds->s.value;
+	else if (role->bounds != bounds->s.value) {
+		yyerror2("role %s has inconsistent bounds %s/%s",
+			 role_id, bounds_id, policydbp->p_role_val_to_name[role->bounds - 1]);
+		return -1;
+	}
+	free(bounds_id);
+
+	return 0;
+}
+
 role_datum_t *declare_role(void)
 {
 	char *id = queue_remove(id_queue), *dest_id = NULL;
@@ -206,6 +245,12 @@ role_datum_t *declare_role(void)
 			}
 			role_datum_init(dest_role);
 			dest_role->s.value = value;
+			if (role_implicit_bounds(roles_tab, dest_id, dest_role)) {
+				free(dest_id);
+				role_datum_destroy(dest_role);
+				free(dest_role);
+				return NULL;
+			}
 			if (hashtab_insert(roles_tab, dest_id, dest_role)) {
 				yyerror("Out of memory!");
 				free(dest_id);
@@ -221,30 +266,30 @@ role_datum_t *declare_role(void)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return NULL;
-		}
+		yyerror("Out of memory!");
+		return NULL;
+	}
 	case -2:{
-			yyerror("duplicate declaration of role");
-			return NULL;
-		}
+		yyerror("duplicate declaration of role");
+		return NULL;
+	}
 	case -1:{
-			yyerror("could not declare role here");
+		yyerror("could not declare role here");
+		return NULL;
+	}
+	case 0:{
+		if (ebitmap_set_bit(&dest_role->dominates, role->s.value - 1, 1)) {
+			yyerror("out of memory");
 			return NULL;
 		}
-	case 0:{
-			if (ebitmap_set_bit(&dest_role->dominates, role->s.value - 1, 1)) {
-				yyerror("out of memory");
-				return NULL;
-			}
-			return dest_role;
-		}
+		return dest_role;
+	}
 	case 1:{
-			return dest_role;	/* role already declared for this block */
-		}
+		return dest_role;      /* role already declared for this block */
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 }
 
@@ -289,25 +334,59 @@ type_datum_t *declare_type(unsigned char primary, unsigned char isattr)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return NULL;
-		}
+		yyerror("Out of memory!");
+		return NULL;
+	}
 	case -2:{
-			yyerror2("duplicate declaration of type/attribute");
-			return NULL;
-		}
+		yyerror2("duplicate declaration of type/attribute");
+		return NULL;
+	}
 	case -1:{
-			yyerror("could not declare type/attribute here");
-			return NULL;
-		}
+		yyerror("could not declare type/attribute here");
+		return NULL;
+	}
 	case 0:
 	case 1:{
-			return typdatum;
-		}
-	default:{
-			assert(0);     /* should never get here */
-		}
+		return typdatum;
 	}
+	default:{
+		assert(0);	       /* should never get here */
+	}
+	}
+}
+
+static int user_implicit_bounds(hashtab_t users_tab, char *user_id, user_datum_t * user)
+{
+	user_datum_t *bounds;
+	char *bounds_id, *delim;
+
+	delim = strrchr(user_id, '.');
+	if (!delim)
+		return 0;	       /* no implicit boundary */
+
+	bounds_id = strdup(user_id);
+	if (!bounds_id) {
+		yyerror("out of memory");
+		return -1;
+	}
+	bounds_id[(size_t) (delim - user_id)] = '\0';
+
+	bounds = hashtab_search(users_tab, bounds_id);
+	if (!bounds) {
+		yyerror2("user %s doesn't exist, is implicit bounds of %s", bounds_id, user_id);
+		return -1;
+	}
+
+	if (!user->bounds)
+		user->bounds = bounds->s.value;
+	else if (user->bounds != bounds->s.value) {
+		yyerror2("user %s has inconsistent bounds %s/%s",
+			 user_id, bounds_id, policydbp->p_role_val_to_name[user->bounds - 1]);
+		return -1;
+	}
+	free(bounds_id);
+
+	return 0;
 }
 
 user_datum_t *declare_user(void)
@@ -361,6 +440,12 @@ user_datum_t *declare_user(void)
 			}
 			user_datum_init(dest_user);
 			dest_user->s.value = value;
+			if (user_implicit_bounds(users_tab, dest_id, dest_user)) {
+				free(dest_id);
+				user_datum_destroy(dest_user);
+				free(dest_user);
+				return NULL;
+			}
 			if (hashtab_insert(users_tab, dest_id, dest_user)) {
 				yyerror("Out of memory!");
 				free(dest_id);
@@ -376,26 +461,26 @@ user_datum_t *declare_user(void)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return NULL;
-		}
+		yyerror("Out of memory!");
+		return NULL;
+	}
 	case -2:{
-			yyerror("duplicate declaration of user");
-			return NULL;
-		}
+		yyerror("duplicate declaration of user");
+		return NULL;
+	}
 	case -1:{
-			yyerror("could not declare user here");
-			return NULL;
-		}
+		yyerror("could not declare user here");
+		return NULL;
+	}
 	case 0:{
-			return dest_user;
-		}
+		return dest_user;
+	}
 	case 1:{
-			return dest_user;	/* user already declared for this block */
-		}
+		return dest_user;      /* user already declared for this block */
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 }
 
@@ -479,7 +564,12 @@ int require_symbol(uint32_t symbol_type, hashtab_key_t key, hashtab_datum_t datu
 		symtab_datum_t *s = (symtab_datum_t *) hashtab_search(policydbp->symtab[symbol_type].table,
 								      key);
 		assert(s != NULL);
-		*dest_value = s->value;
+
+		if (symbol_type == SYM_LEVELS) {
+			*dest_value = ((level_datum_t *) s)->level->sens;
+		} else {
+			*dest_value = s->value;
+		}
 	} else if (retval == -2) {
 		/* ignore require statements if that symbol was
 		 * previously declared and is in current scope */
@@ -593,41 +683,41 @@ int require_class(int pass)
 	ret = require_symbol(SYM_CLASSES, class_id, datum, &datum->s.value, &datum->s.value);
 	switch (ret) {
 	case -3:{
-			yyerror("Out of memory!");
-			free(class_id);
-			class_datum_destroy(datum);
-			goto cleanup;
-		}
+		yyerror("Out of memory!");
+		free(class_id);
+		class_datum_destroy(datum);
+		goto cleanup;
+	}
 	case -2:{
-			yyerror("duplicate declaration of class");
-			free(class_id);
-			class_datum_destroy(datum);
-			goto cleanup;
-		}
+		yyerror("duplicate declaration of class");
+		free(class_id);
+		class_datum_destroy(datum);
+		goto cleanup;
+	}
 	case -1:{
-			yyerror("could not require class here");
-			free(class_id);
-			class_datum_destroy(datum);
+		yyerror("could not require class here");
+		free(class_id);
+		class_datum_destroy(datum);
+		goto cleanup;
+	}
+	case 0:{
+		/* a new class was added; reindex everything */
+		if (policydb_index_classes(policydbp)) {
+			yyerror("Out of memory!");
 			goto cleanup;
 		}
-	case 0:{
-			/* a new class was added; reindex everything */
-			if (policydb_index_classes(policydbp)) {
-				yyerror("Out of memory!");
-				goto cleanup;
-			}
-			break;
-		}
+		break;
+	}
 	case 1:{
-			class_datum_destroy(datum);
-			datum = hashtab_search(policydbp->p_classes.table, class_id);
-			assert(datum); /* the class datum should have existed */
-			free(class_id);
-			break;
-		}
+		class_datum_destroy(datum);
+		datum = hashtab_search(policydbp->p_classes.table, class_id);
+		assert(datum);	       /* the class datum should have existed */
+		free(class_id);
+		break;
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 
 	/* now add each of the permissions to this class's requirements */
@@ -706,31 +796,31 @@ int require_role(int pass)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return -1;
-		}
+		yyerror("Out of memory!");
+		return -1;
+	}
 	case -2:{
-			yyerror("duplicate declaration of role");
-			return -1;
-		}
+		yyerror("duplicate declaration of role");
+		return -1;
+	}
 	case -1:{
-			yyerror("could not require role here");
+		yyerror("could not require role here");
+		return -1;
+	}
+	case 0:{
+		/* all roles dominate themselves */
+		if (ebitmap_set_bit(&role->dominates, role->s.value - 1, 1)) {
+			yyerror("Out of memory");
 			return -1;
 		}
-	case 0:{
-			/* all roles dominate themselves */
-			if (ebitmap_set_bit(&role->dominates, role->s.value - 1, 1)) {
-				yyerror("Out of memory");
-				return -1;
-			}
-			return 0;
-		}
+		return 0;
+	}
 	case 1:{
-			return 0;      /* role already required */
-		}
+		return 0;	       /* role already required */
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 }
 
@@ -762,26 +852,26 @@ static int require_type_or_attribute(int pass, unsigned char isattr)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return -1;
-		}
+		yyerror("Out of memory!");
+		return -1;
+	}
 	case -2:{
-			yyerror("duplicate declaration of type/attribute");
-			return -1;
-		}
+		yyerror("duplicate declaration of type/attribute");
+		return -1;
+	}
 	case -1:{
-			yyerror("could not require type/attribute here");
-			return -1;
-		}
+		yyerror("could not require type/attribute here");
+		return -1;
+	}
 	case 0:{
-			return 0;
-		}
+		return 0;
+	}
 	case 1:{
-			return 0;      /* type already required */
-		}
+		return 0;	       /* type already required */
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 }
 
@@ -821,26 +911,26 @@ int require_user(int pass)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return -1;
-		}
+		yyerror("Out of memory!");
+		return -1;
+	}
 	case -2:{
-			yyerror("duplicate declaration of user");
-			return -1;
-		}
+		yyerror("duplicate declaration of user");
+		return -1;
+	}
 	case -1:{
-			yyerror("could not require user here");
-			return -1;
-		}
+		yyerror("could not require user here");
+		return -1;
+	}
 	case 0:{
-			return 0;
-		}
+		return 0;
+	}
 	case 1:{
-			return 0;      /* user already required */
-		}
+		return 0;	       /* user already required */
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 }
 
@@ -868,26 +958,26 @@ int require_bool(int pass)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return -1;
-		}
+		yyerror("Out of memory!");
+		return -1;
+	}
 	case -2:{
-			yyerror("duplicate declaration of boolean");
-			return -1;
-		}
+		yyerror("duplicate declaration of boolean");
+		return -1;
+	}
 	case -1:{
-			yyerror("could not require boolean here");
-			return -1;
-		}
+		yyerror("could not require boolean here");
+		return -1;
+	}
 	case 0:{
-			return 0;
-		}
+		return 0;
+	}
 	case 1:{
-			return 0;      /* boolean already required */
-		}
+		return 0;	       /* boolean already required */
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 }
 
@@ -930,26 +1020,26 @@ int require_sens(int pass)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return -1;
-		}
+		yyerror("Out of memory!");
+		return -1;
+	}
 	case -2:{
-			yyerror("duplicate declaration of sensitivity");
-			return -1;
-		}
+		yyerror("duplicate declaration of sensitivity");
+		return -1;
+	}
 	case -1:{
-			yyerror("could not require sensitivity here");
-			return -1;
-		}
+		yyerror("could not require sensitivity here");
+		return -1;
+	}
 	case 0:{
-			return 0;
-		}
+		return 0;
+	}
 	case 1:{
-			return 0;      /* sensitivity already required */
-		}
+		return 0;	       /* sensitivity already required */
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 }
 
@@ -982,26 +1072,26 @@ int require_cat(int pass)
 	}
 	switch (retval) {
 	case -3:{
-			yyerror("Out of memory!");
-			return -1;
-		}
+		yyerror("Out of memory!");
+		return -1;
+	}
 	case -2:{
-			yyerror("duplicate declaration of category");
-			return -1;
-		}
+		yyerror("duplicate declaration of category");
+		return -1;
+	}
 	case -1:{
-			yyerror("could not require category here");
-			return -1;
-		}
+		yyerror("could not require category here");
+		return -1;
+	}
 	case 0:{
-			return 0;
-		}
+		return 0;
+	}
 	case 1:{
-			return 0;      /* category already required */
-		}
+		return 0;	       /* category already required */
+	}
 	default:{
-			assert(0);     /* should never get here */
-		}
+		assert(0);	       /* should never get here */
+	}
 	}
 }
 
@@ -1220,7 +1310,7 @@ int begin_optional_else(int pass)
 		stack_top->decl->next = decl;
 	} else {
 		/* pick the (hopefully last) declaration of this
-		 * avrule block, built from pass 1 */
+		   avrule block, built from pass 1 */
 		decl = stack_top->decl->next;
 		assert(decl != NULL && decl->next == NULL && decl->decl_id == next_decl_id);
 	}
@@ -1316,14 +1406,14 @@ static int push_stack(int stack_type, ...)
 	va_start(ap, stack_type);
 	switch (s->type = stack_type) {
 	case 1:{
-			s->u.avrule = va_arg(ap, avrule_block_t *);
-			s->decl = va_arg(ap, avrule_decl_t *);
-			break;
-		}
+		s->u.avrule = va_arg(ap, avrule_block_t *);
+		s->decl = va_arg(ap, avrule_decl_t *);
+		break;
+	}
 	case 2:{
-			s->u.cond_list = va_arg(ap, cond_list_t *);
-			break;
-		}
+		s->u.cond_list = va_arg(ap, cond_list_t *);
+		break;
+	}
 	default:
 		/* invalid stack type given */
 		assert(0);
