@@ -5,6 +5,7 @@
  *  @author Jeremy A. Mowery jmowery@tresys.com
  *  @author Jason Tang jtang@tresys.com
  *  @author Brandon Whalen bwhalen@tresys.com
+ *  @author Jeremy Solt jsolt@tresys.com
  *
  *  Copyright (C) 2006-2008 Tresys Technology, LLC
  *
@@ -45,6 +46,7 @@
 #include <sepol/policydb.h>
 #include <sepol/module.h>
 #include <sepol/policydb/module.h>
+#include <sepol/policydb/avrule_block.h>
 
 #include <stdbool.h>
 #include <qpol/iterator.h>
@@ -186,11 +188,12 @@ static int read_source_policy(qpol_policy_t * qpolicy, char *progname, int optio
 	INFO(qpolicy, "%s", "Parsing policy. (Step 1 of 5)");
 	init_scanner();
 	init_parser(1, load_rules);
+	errno = 0;
 	if (yyparse() || policydb_errors) {
 		ERR(qpolicy, "%s:  error(s) encountered while parsing configuration\n", progname);
 		queue_destroy(id_queue);
 		id_queue = NULL;
-		errno = EIO;
+//		errno = EIO;
 		return -1;
 	}
 	/* rewind the pointer */
@@ -201,13 +204,13 @@ static int read_source_policy(qpol_policy_t * qpolicy, char *progname, int optio
 		ERR(qpolicy, "%s:  error(s) encountered while parsing configuration\n", progname);
 		queue_destroy(id_queue);
 		id_queue = NULL;
-		errno = EIO;
+//		errno = EIO;
 		return -1;
 	}
 	queue_destroy(id_queue);
 	id_queue = NULL;
 	if (policydb_errors) {
-		errno = EIO;
+//		errno = EIO;
 		return -1;
 	}
 	return 0;
@@ -334,6 +337,20 @@ int qpol_is_file_binpol(FILE * fp)
 	return rt;
 }
 
+int qpol_is_data_mod_pkg(char * data)
+{
+	size_t sz;
+	__u32 ubuf;
+
+	memcpy(&ubuf, data, sizeof(__u32));
+
+	ubuf = le32_to_cpu(ubuf);
+	if (ubuf == SEPOL_MODULE_PACKAGE_MAGIC)
+		return 1;
+
+	return 0;
+}
+
 int qpol_is_file_mod_pkg(FILE * fp)
 {
 	size_t sz;
@@ -470,6 +487,241 @@ static int infer_policy_version(qpol_policy_t * policy)
 	return STATUS_SUCCESS;
 }
 
+/** State tracking struct used in the functions check_disabled, remove_symbol, and prune_disabled_symbols to handle disabled symbols */
+struct symbol_pruning_state
+{
+	qpol_policy_t *p; /**< The policy */
+	int symbol_type; /**< The current symbol type being processed */
+};
+
+/** Apply callback for hashtab_map_remove_on_error.
+ *  This function tests whether a symbol referenced by the policy is declared or only ever required.
+ *  Symbols without a declaration are disabled and must be removed.
+ *  @param key Symbol key to check.
+ *  @param datum Symbol datum to check.
+ *  @param args State object (of type struct symbol_pruning_state)
+ *  @return 0 if symbol is enabled, 1 if not enabled.
+ */
+static int check_disabled(hashtab_key_t key, hashtab_datum_t datum, void *args)
+{
+	struct symbol_pruning_state *s = args;
+	if (!is_id_enabled((char *)key, &(s->p->p->p), s->symbol_type))
+		return 1;
+	return 0;
+}
+
+/** Remove callback for hashtab_map_remove_on_error.
+ *  Frees all memory associated with a disabled symbol that has been removed from the symbol table.
+ *  @param key Symbol key to remove
+ *  @param datum Symbol datum to remove
+ *  @param args State object (of type struct symbol_pruning_state)
+ *  @post All memory associated with the symbol is freed.
+ */
+static void remove_symbol(hashtab_key_t key, hashtab_datum_t datum, void *args)
+{
+	struct symbol_pruning_state *s = args;
+	switch (s->symbol_type) {
+	case SYM_ROLES:
+	{
+		role_datum_destroy((role_datum_t *) datum);
+		break;
+	}
+	case SYM_TYPES:
+	{
+		type_datum_destroy((type_datum_t *) datum);
+		break;
+	}
+	case SYM_USERS:
+	{
+		user_datum_destroy((user_datum_t *) datum);
+		break;
+	}
+	case SYM_BOOLS:
+	{
+		/* no-op */
+		break;
+	}
+	case SYM_LEVELS:
+	{
+		level_datum_destroy((level_datum_t *) datum);
+		break;
+	}
+	case SYM_CATS:
+	{
+		cat_datum_destroy((cat_datum_t *) datum);
+		break;
+	}
+	default:
+		return;		       /* invalid type of datum to free; do nothing */
+	}
+	free(key);
+	free(datum);
+}
+
+/** Remove symbols that are only required but never declared from the policy.
+ *  Removes each disabled symbol freeing all memory associated with it.
+ *  @param policy The policy from which disabled symbols should be removed.
+ *  @return always 0.
+ *  @note Since hashtab_map_remove_on_error does not return any error status,
+ *  it is impossible to tell if it has failed; if it fails, the policy will
+ *  be in an inconsistent state.
+ */
+static int prune_disabled_symbols(qpol_policy_t * policy)
+{
+	if (policy->type == QPOL_POLICY_KERNEL_BINARY)
+		return 0;	       /* checkpolicy already prunes disabled symbols */
+	struct symbol_pruning_state state;
+	state.p = policy;
+	for (state.symbol_type = SYM_ROLES; state.symbol_type < SYM_NUM; state.symbol_type++) {
+		hashtab_map_remove_on_error(policy->p->p.symtab[state.symbol_type].table, check_disabled, remove_symbol, &state);
+	}
+	return 0;
+}
+
+/** For all symbols that are multiply defined (such as attributes, roles, and users),
+ *  union the relevant sets of types and roles from each declaration.
+ *  @param policy The policy containig the symbols to union.
+ *  @return 0 on success, non-zero on error; if the call fails,
+ *  errno will be set, and the policy should be considered invalid.
+ */
+static int union_multiply_declared_symbols(qpol_policy_t * policy) {
+	/* general structure of this function:
+	walk role and user symbol tables for each role/user/attribute
+		get datum from symtab, get key from array
+		look up symbol in scope table
+		foreach decl_id in scope entry
+			union types/roles bitmap with datum's copy
+	*/
+	qpol_iterator_t * iter = NULL;
+	int error = 0;
+	if (qpol_policy_get_type_iter(policy, &iter)) {
+		return 1;
+	}
+	for (; !qpol_iterator_end(iter); qpol_iterator_next(iter)) {
+		type_datum_t *attr;
+		if (qpol_iterator_get_item(iter, (void**)&attr)) {
+			error = errno;
+			goto err;
+		}
+		unsigned char isattr = 0;
+		if (qpol_type_get_isattr(policy, attr, &isattr)) {
+			error = errno;
+			goto err;
+		}
+		if (!isattr)
+			continue;
+		const char *name;
+		if (qpol_type_get_name(policy, (qpol_type_t*)attr, &name)) {
+			error = errno;
+			goto err;
+		}
+		policydb_t *db = &policy->p->p;
+		avrule_block_t *blk = db->global;
+		for (; blk; blk = blk->next) {
+			avrule_decl_t *decl = blk->enabled;
+			if (!decl)
+				continue; /* disabled */
+			type_datum_t *internal_datum = hashtab_search(decl->symtab[SYM_TYPES].table, (const hashtab_key_t)name);
+			if (internal_datum == NULL) {
+				continue; /* not declared here */
+			}
+			if (ebitmap_union(&attr->types, &internal_datum->types))
+			{
+				error = errno;
+				ERR(policy, "could not merge declarations for attribute %s", name);
+				goto err;
+			}
+		}
+	}
+	qpol_iterator_destroy(&iter);
+
+	/* repeat for roles */
+	if (qpol_policy_get_role_iter(policy, &iter)) {
+		return 1;
+	}
+	for (; !qpol_iterator_end(iter); qpol_iterator_next(iter)) {
+		role_datum_t *role;
+		if (qpol_iterator_get_item(iter, (void**)&role)) {
+			error = errno;
+			goto err;
+		}
+		const char *name;
+		if (qpol_role_get_name(policy, (qpol_role_t*)role, &name)) {
+			error = errno;
+			goto err;
+		}
+		policydb_t *db = &policy->p->p;
+		scope_datum_t* scope_datum = hashtab_search(db->scope[SYM_ROLES].table, (const hashtab_key_t)name);
+		if (scope_datum == NULL) {
+			ERR(policy, "could not find scope datum for role %s", name);
+			error = ENOENT;
+			goto err;
+		}
+		for (uint32_t i = 0; i < scope_datum->decl_ids_len; i++)
+		{
+			if (db->decl_val_to_struct[scope_datum->decl_ids[i] - 1]->enabled == 0)
+				continue; /* block is disabled */
+			role_datum_t *internal_datum = hashtab_search(db->decl_val_to_struct[scope_datum->decl_ids[i] - 1]->symtab[SYM_ROLES].table, (const hashtab_key_t)name);
+			if (internal_datum == NULL) {
+				continue; /* not declared here */
+			}
+			if (ebitmap_union(&role->types.types, &internal_datum->types.types) || ebitmap_union(&role->dominates, &internal_datum->dominates))
+			{
+				error = errno;
+				ERR(policy, "could not merge declarations for role %s", name);
+				goto err;
+			}
+		}
+	}
+	qpol_iterator_destroy(&iter);
+
+	/* repeat for users */
+	if (qpol_policy_get_user_iter(policy, &iter)) {
+		return 1;
+	}
+	for (; !qpol_iterator_end(iter); qpol_iterator_next(iter)) {
+		user_datum_t *user;
+		if (qpol_iterator_get_item(iter, (void**)&user)) {
+			error = errno;
+			goto err;
+		}
+		const char *name;
+		if (qpol_user_get_name(policy, (qpol_user_t*)user, &name)) {
+			error = errno;
+			goto err;
+		}
+		policydb_t *db = &policy->p->p;
+		scope_datum_t* scope_datum = hashtab_search(db->scope[SYM_USERS].table, (const hashtab_key_t)name);
+		if (scope_datum == NULL) {
+			ERR(policy, "could not find scope datum for user %s", name);
+			error = ENOENT;
+			goto err;
+		}
+		for (uint32_t i = 0; i < scope_datum->decl_ids_len; i++)
+		{
+			if (db->decl_val_to_struct[scope_datum->decl_ids[i] - 1]->enabled == 0)
+				continue; /* block is disabled */
+			user_datum_t *internal_datum = hashtab_search(db->decl_val_to_struct[scope_datum->decl_ids[i] -1 ]->symtab[SYM_USERS].table, (const hashtab_key_t)name);
+			if (internal_datum == NULL) {
+				continue; /* not declared here */
+			}
+			if (ebitmap_union(&user->roles.roles, &internal_datum->roles.roles))
+			{
+				error = errno;
+				ERR(policy, "could not merge declarations for user %s", name);
+				goto err;
+			}
+		}
+	}
+	qpol_iterator_destroy(&iter);
+
+	return 0;
+err:
+	qpol_iterator_destroy(&iter);
+	errno = error;
+	return 1;
+}
+
 /* forward declarations see policy_extend.c */
 struct qpol_extended_image;
 extern void qpol_extended_image_destroy(struct qpol_extended_image **ext);
@@ -496,7 +748,7 @@ int qpol_policy_rebuild_opt(qpol_policy_t * policy, const int options)
 	sepol_policydb_t **modules = NULL;
 	qpol_module_t *base = NULL;
 	size_t num_modules = 0, i;
-	int error = 0, old_otions;
+	int error = 0, old_options;
 
 	if (!policy) {
 		ERR(NULL, "%s", strerror(EINVAL));
@@ -517,7 +769,7 @@ int qpol_policy_rebuild_opt(qpol_policy_t * policy, const int options)
 	policy->p = NULL;
 	struct qpol_extended_image *ext = policy->ext;
 	policy->ext = NULL;
-	old_otions = policy->options;
+	old_options = policy->options;
 	policy->options = options;
 
 	/* QPOL_POLICY_OPTION_NO_RULES implies QPOL_POLICY_OPTION_NO_NEVERALLOWS */
@@ -583,7 +835,17 @@ int qpol_policy_rebuild_opt(qpol_policy_t * policy, const int options)
 		avtab_init(&(policy->p->p.te_cond_avtab));
 	}
 
-	if (qpol_expand_module(policy, !(options & (QPOL_POLICY_OPTION_NO_NEVERALLOWS)))) {
+	if (prune_disabled_symbols(policy)) {
+		error = errno;
+		goto err;
+	}
+
+	if (union_multiply_declared_symbols(policy)) {
+		error = errno;
+		goto err;
+	}
+
+	if (qpol_expand_module(policy, !(policy->options & (QPOL_POLICY_OPTION_NO_NEVERALLOWS)))) {
 		error = errno;
 		goto err;
 	}
@@ -608,7 +870,7 @@ int qpol_policy_rebuild_opt(qpol_policy_t * policy, const int options)
 
 	policy->p = old_p;
 	policy->ext = ext;
-	policy->options = old_otions;
+	policy->options = old_options;
 	errno = error;
 	return STATUS_ERR;
 }
@@ -672,6 +934,7 @@ int qpol_policy_open_from_file_opt(const char *path, qpol_policy_t ** policy, qp
 		return -1;
 	}
 
+    errno = 0;
 	if (!(*policy = calloc(1, sizeof(qpol_policy_t)))) {
 		error = errno;
 		ERR(NULL, "%s", strerror(error));
@@ -717,11 +980,12 @@ int qpol_policy_open_from_file_opt(const char *path, qpol_policy_t ** policy, qp
 
 	sepol_policy_file_set_handle(pfile, (*policy)->sh);
 
+    errno=0;
 	if (qpol_is_file_binpol(infile)) {
 		(*policy)->type = retv = QPOL_POLICY_KERNEL_BINARY;
 		sepol_policy_file_set_fp(pfile, infile);
 		if (sepol_policydb_read((*policy)->p, pfile)) {
-			error = EIO;
+//			error = EIO;
 			goto err;
 		}
 		/* By definition, binary policy cannot have neverallow rules and all other rules are always loaded. */
@@ -731,13 +995,9 @@ int qpol_policy_open_from_file_opt(const char *path, qpol_policy_t ** policy, qp
 			error = errno;
 			goto err;
 		}
-	} else if (qpol_is_file_mod_pkg(infile)) {
+	} else if (qpol_module_create_from_file(path, &mod) == STATUS_SUCCESS) {
 		(*policy)->type = retv = QPOL_POLICY_MODULE_BINARY;
-		if (qpol_module_create_from_file(path, &mod)) {
-			error = errno;
-			ERR(*policy, "%s", strerror(error));
-			goto err;
-		}
+
 		if (qpol_policy_append_module(*policy, mod)) {
 			error = errno;
 			goto err;
@@ -793,6 +1053,16 @@ int qpol_policy_open_from_file_opt(const char *path, qpol_policy_t ** policy, qp
 		avtab_init(&((*policy)->p->p.te_avtab));
 		avtab_init(&((*policy)->p->p.te_cond_avtab));
 
+		if (prune_disabled_symbols(*policy)) {
+			error = errno;
+			goto err;
+		}
+
+		if (union_multiply_declared_symbols(*policy)) {
+			error = errno;
+			goto err;
+		}
+
 		/* expand */
 		if (qpol_expand_module(*policy, !(options & (QPOL_POLICY_OPTION_NO_NEVERALLOWS)))) {
 			error = errno;
@@ -911,7 +1181,17 @@ int qpol_policy_open_from_memory_opt(qpol_policy_t ** policy, const char *fileda
 	avtab_init(&((*policy)->p->p.te_avtab));
 	avtab_init(&((*policy)->p->p.te_cond_avtab));
 
-	/* expand :) */
+	if (prune_disabled_symbols(*policy)) {
+		error = errno;
+		goto err;
+	}
+
+	if (union_multiply_declared_symbols(*policy)) {
+		error = errno;
+		goto err;
+	}
+
+	/* expand */
 	if (qpol_expand_module(*policy, !(options & (QPOL_POLICY_OPTION_NO_NEVERALLOWS)))) {
 		error = errno;
 		goto err;
@@ -1214,7 +1494,7 @@ int qpol_policy_has_capability(const qpol_policy_t * policy, qpol_capability_e c
 	switch (cap) {
 	case QPOL_CAP_ATTRIB_NAMES:
 	{
-		if (policy->type == QPOL_POLICY_KERNEL_SOURCE || policy->type == QPOL_POLICY_MODULE_BINARY)
+		if ((policy->type == QPOL_POLICY_KERNEL_SOURCE || policy->type == QPOL_POLICY_MODULE_BINARY) || (version >= 24))
 			return 1;
 		break;
 	}
